#version 430 core
/*
* Copyright 2023 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the GCG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	vec4 ShadowPosition;
} vert;

out vec4 color;

uniform vec3 camera_world;


uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;
uniform sampler2D metalTexture;
uniform sampler2D normalTexture;
uniform sampler2D aoTexture;
uniform sampler2D shadowMap;

uniform vec3 lightPosition;

uniform struct LightSource {
	vec3 position;
	vec3 color;
	vec3 attenuation;
} dirL;

uniform struct LightSource {
	vec3 position;
	vec3 color;
	vec3 attenuation;
} pointL1;

uniform struct LightSource {
	vec3 position;
	vec3 color;
	vec3 attenuation;
} pointL2;

uniform struct LightSource {
	vec3 position;
	vec3 color;
	vec3 attenuation;
} pointL3;

const float PI = 3.14159265359;


vec3 getNormalFromMap(vec3 Normal)
{
    vec3 tangentNormal = texture(normalTexture, vert.uv).rgb * 2.0 - 1.0;

    vec3 Q1  = dFdx(vert.position_world);
    vec3 Q2  = dFdy(vert.position_world);
    vec2 st1 = dFdx(vert.uv);
    vec2 st2 = dFdy(vert.uv);

    vec3 N   = normalize(Normal);
    vec3 T  = normalize(Q1*st2.t - Q2*st1.t);
    vec3 B  = -normalize(cross(N, T));
    mat3 TBN = mat3(T, B, N);

    return normalize(TBN * tangentNormal);
}

// Computes the reflection direction for an incident vector I about normal N, 
// and clamps the reflection to a maximum of 180, i.e. the reflection vector
// will always lie within the hemisphere around normal N.
// Aside from clamping, this function produces the same result as GLSL's reflect function.
vec3 clampedReflect(vec3 I, vec3 N)
{
	return I - 2.0 * min(dot(N, I), 0.0) * N;
}

vec3 fresnelSchlick (float cosTheta, vec3 F0) {
    return F0 + (1.0 - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a = roughness*roughness;
    float a2 = a*a;
    float NdotH = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float nom   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return nom / denom;
}


float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float nom   = NdotV;
    float denom = NdotV * (1.0 - k) + k;

    return nom / denom;
}


float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2 = GeometrySchlickGGX(NdotV, roughness);
    float ggx1 = GeometrySchlickGGX(NdotL, roughness);

    return ggx1 * ggx2;
}

vec3 calcIntegral(LightSource lgt, vec3 V, vec3 N, vec3 baseCol, float roughness, float metalness,  vec3 F0)
{
	// Step 1. Calculate Radiance of fragment.
	vec3 lengthVec = lgt.position - vert.position_world;
	
	vec3 L = normalize(lengthVec);
	vec3 H = normalize(V + L);

	float dist = length(lengthVec);
	vec3 radiance = lgt.color * lgt.attenuation;

	// Step 2. Cook-Torrence BRDF
	float NDF = DistributionGGX(N, H, roughness);
	float G = GeometrySmith(N, V, L, roughness);
	vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);

	// Step 3. calculate specular part
	vec3 numerator    = NDF * G * F; 
    float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001; // + 0.0001 to prevent divide by zero
    vec3 specular = numerator / denominator;

	// Step 4. calculate kS, kD and other necessary components for L0
	vec3 kS = F;

	vec3 kD = vec3(1.0) - kS;
	kD *= 1.0 - metalness;

	float NdotL = max(dot(N, L), 0.0);  
	
	
	return (kD * baseCol / PI + specular) * radiance * NdotL;
}

vec3 calcIntegralForDirectional(LightSource lgt, vec3 V, vec3 N, vec3 baseCol, float roughness, float metalness,  vec3 F0)
{
	// Step 1. Calculate Radiance of fragment.
	vec3 L = normalize(lgt.position);
	
	vec3 H = normalize(V + L);

	float dist = 1;
	vec3 radiance = lgt.color * lgt.attenuation;

	// Step 2. Cook-Torrence BRDF
	float NDF = DistributionGGX(N, H, roughness);
	float G = GeometrySmith(N, V, L, roughness);
	vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);

	// Step 3. calculate specular part
	vec3 numerator    = NDF * G * F; 
    float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001; // + 0.0001 to prevent divide by zero
    vec3 specular = numerator / denominator;

	// Step 4. calculate kS, kD and other necessary components for L0
	vec3 kS = F;

	vec3 kD = vec3(1.0) - kS;
	kD *= 1.0 - metalness;

	float NdotL = max(dot(N, L), 0.0);  
	
	
	return (kD * baseCol / PI + specular) * radiance * NdotL;
}


float ShadowCalculation(vec4 fragPosLightSpace, vec3 n)
{
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
    projCoords = projCoords * 0.5 + 0.5;

	float dotProd = dot(n, lightPosition);
    if(dotProd < 0){
       return 0.5f;
    }

    float currentDepth = projCoords.z;
    float bias = max(0.001 * (1.0 - dotProd), 0.0024);



    float shadow = 0.0;
    vec2 texelSize = 1.0 / textureSize(shadowMap, 0);

    for(int x = -1; x <= 1; x++)
    {
        for(int y = -1; y <= 1; y++)
        {
            float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r; 
            shadow += currentDepth - bias > pcfDepth ? 0.5 : 0.0;
        }
    }

    shadow /= 9.0;
    
    return 1.0 - shadow;
}

void main() {	
	
	vec3 n = normalize(vert.normal_world);
	n = getNormalFromMap(n);


	//static PBR attributes, should be replaced by TexMaps later
	vec3 baseCol = texture(diffuseTexture, vert.uv).rgb;
	float ao = texture(aoTexture, vert.uv).r;
	float roughness = texture(specularTexture, vert.uv).r;
	float metallic = texture(metalTexture, vert.uv).r;


	//Basic variables needed
		
	vec3 v = normalize(camera_world - vert.position_world);
	vec3 R = normalize(clampedReflect(v, n));


	vec3 F0 = vec3(0.04); // <-- some kind of plastic
	F0 = mix(F0, baseCol, metallic);

	vec3 Lo = vec3(0.0);

	// add all light from lightsources
	Lo += calcIntegralForDirectional(dirL, v, n, baseCol, roughness, metallic, F0);
	Lo += calcIntegral(pointL1, v, n, baseCol, roughness, metallic, F0);
	Lo += calcIntegral(pointL2, v, n, baseCol, roughness, metallic, F0);
	Lo += calcIntegral(pointL3, v, n, baseCol, roughness, metallic, F0);


	// add in ambient light (replace with environment light)
	vec3 ambient = vec3(0.00) * baseCol * ao;
	Lo += ambient;
		
	//Add shadow for directional light.
	float shadow = ShadowCalculation(vert.ShadowPosition, n);
	Lo *= shadow;

	// HDR correction and gamma correction
	Lo = Lo / (Lo + vec3(1.0f));
	Lo = pow(Lo, vec3(1.0/2.2)); 
	color = vec4(Lo, 1.0f);
		

	
}

